/*
 * Copyright © 2020-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of gaming-core.
 *
 * Gaming-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Gaming-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * gaming-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.core.domain;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import de.fhdw.gaming.core.domain.util.ConsumerE;

/**
 * Implements {@link Game}.
 *
 * @param <S>  The type of game states.
 * @param <M>  The type of game moves.
 * @param <P>  The type of game players.
 * @param <ST> The type of game strategies.
 */
@SuppressWarnings("PMD.GodClass")
public final class DefaultGame<P extends Player<P>, S extends State<P, S>, M extends Move<P, S>,
        ST extends Strategy<P, S, M>>
        implements Game<P, S, M, ST> {

    /**
     * The ID of this game.
     */
    private final int id;
    /**
     * The game state.
     */
    private S state;
    /**
     * The players of the game together with their strategies.
     */
    private final Map<String, ST> strategies;
    /**
     * The maximum computation time per move in seconds.
     */
    private final long maxComputationTimePerMove;
    /**
     * The move checker.
     */
    private final MoveChecker<P, S, M> moveChecker;
    /**
     * The move generator.
     */
    private final MoveGenerator<P, S, M> moveGenerator;
    /**
     * The registered observers.
     */
    private final List<Observer> observers;
    /**
     * The executor for submitting tasks for choosing a move.
     */
    private final ExecutorService executorService;
    /**
     * {@code true} if the game has been started, else {@code false}.
     */
    private boolean started;

    /**
     * Creates a game. Uses
     * {@link GameBuilder#DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE} as maximum
     * computation time per move.
     *
     * @param id            The ID of this game.
     * @param initialState  The initial state of the game.
     * @param strategies    The players' strategies.
     * @param moveChecker   The move checker.
     * @param moveGenerator The move generator used for generating a valid but random move.
     * @param observers     The {@link Observer}s to attach.
     * @throws IllegalArgumentException if the player sets do not match.
     * @throws InterruptedException     if creating the game has been interrupted.
     */
    public DefaultGame(final int id, final S initialState, final Map<String, ST> strategies,
            final MoveChecker<P, S, M> moveChecker, final MoveGenerator<P, S, M> moveGenerator,
            final List<Observer> observers)
            throws IllegalArgumentException, InterruptedException {

        this(id, initialState, strategies, GameBuilder.DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE, moveChecker,
                moveGenerator,
                observers);
    }

    /**
     * Creates a game.
     *
     * @param id                        The ID of this game.
     * @param initialState              The initial state of the game.
     * @param strategies                The players' strategies.
     * @param maxComputationTimePerMove The maximum computation time per move in
     *                                  seconds.
     * @param moveChecker               The move checker.
     * @param moveGenerator             The move generator used for generating a valid but random move.
     * @param observers                 The {@link Observer}s to attach.
     * @throws IllegalArgumentException if the player sets do not match.
     * @throws InterruptedException     if creating the game has been interrupted.
     */
    public DefaultGame(final int id, final S initialState, final Map<String, ST> strategies,
            final long maxComputationTimePerMove, final MoveChecker<P, S, M> moveChecker,
            final MoveGenerator<P, S, M> moveGenerator, final List<Observer> observers)
            throws IllegalArgumentException, InterruptedException {

        this.id = id;
        this.state = Objects.requireNonNull(initialState, "initialState").deepCopy();
        this.strategies = new LinkedHashMap<>(Objects.requireNonNull(strategies, "players"));
        this.maxComputationTimePerMove = maxComputationTimePerMove;
        this.moveChecker = Objects.requireNonNull(moveChecker, "moveChecker");
        this.moveGenerator = Objects.requireNonNull(moveGenerator, "moveGenerator");
        this.executorService = Executors.newCachedThreadPool();
        this.started = false;

        if (!strategies.keySet().equals(this.state.getPlayers().keySet())) {
            throw new IllegalArgumentException(
                    "The set of players defined by the game state must match the set of players "
                            + "associated with strategies.");
        }

        this.observers = Collections.synchronizedList(new ArrayList<>(observers));
        this.checkAndAdjustPlayerStatesIfNecessary();
    }

    /**
     * Returns a string representing the state of the game.
     */
    @Override
    public String toString() {
        return String.format("DefaultGame[id=%s, state=%s, strategies=%s]", this.id, this.state, this.strategies);
    }

    @Override
    public int getId() {
        return this.id;
    }

    @Override
    public Map<String, P> getPlayers() {
        return this.state.getPlayers();
    }

    @Override
    public Map<String, ST> getStrategies() {
        return this.strategies;
    }

    @Override
    public S getState() {
        return this.state.deepCopy();
    }

    @Override
    public void addObserver(final Observer observer) {
        this.observers.add(observer);
    }

    @Override
    public void removeObserver(final Observer observer) {
        this.observers.remove(observer);
    }

    @Override
    public void start() throws InterruptedException {
        for (final ST strategy : this.strategies.values()) {
            strategy.reset();
        }
        this.callObservers((final Observer o) -> o.started(this, this.state.deepCopy()));
        this.started = true;
    }

    /**
     * Runs all observers.
     *
     * @param call Called with the observer as argument.
     */
    private void callObservers(final ConsumerE<Observer, InterruptedException> call) throws InterruptedException {
        final ArrayList<Observer> copyOfObserverList = new ArrayList<>(this.observers);
        for (final Observer observer : copyOfObserverList) {
            call.accept(observer);
        }
    }

    @Override
    public void makeMove() throws IllegalStateException, InterruptedException {
        if (!this.isStarted()) {
            throw new IllegalStateException("Trying to make a move although the game has not been started yet.");
        }
        if (this.isFinished()) {
            throw new IllegalStateException("Trying to make a move although the game is already over.");
        }

        final Set<P> nextPlayers = this.state.computeNextPlayers();
        final S stateCopy = this.state.deepCopy();
        final LinkedHashSet<
                P> players = nextPlayers.stream().map((final P player) -> stateCopy.getPlayers().get(player.getName()))
                        .collect(Collectors.toCollection(LinkedHashSet::new));
        this.callObservers((final Observer o) -> o.nextPlayersComputed(this, stateCopy, players));

        if (nextPlayers.isEmpty()) {
            // no active players -> game over
            this.callObservers((final Observer o) -> o.finished(this, this.state.deepCopy()));
            return;
        }

        final CompletionService<Optional<M>> completionService = new ExecutorCompletionService<>(this.executorService);
        final Map<Future<Optional<M>>, P> futures = this.submitMoveComputingRequests(nextPlayers, completionService);
        try {
            this.applyNextPossibleMove(completionService, futures);
        } finally {
            this.state.nextTurn();
            this.checkAndAdjustPlayerStatesIfNecessary();
        }
    }

    @Override
    public void abortRequested() {
        for (final ST strategy : this.strategies.values()) {
            strategy.abortRequested(this.id);
        }
    }

    @Override
    public Optional<M> chooseRandomMove(final P player, final S stateCopy) {
        return this.moveGenerator.generate(player, stateCopy);
    }

    @Override
    public boolean isStarted() {
        return this.started;
    }

    @Override
    public boolean isFinished() {
        final List<P> playersPlaying = this.getPlayers().values().stream()
                .filter((final P player) -> player.getState().equals(PlayerState.PLAYING)).collect(Collectors.toList());
        return playersPlaying.isEmpty();
    }

    @Override
    public void close() {
        this.executorService.shutdown();
    }

    /**
     * Places a move choosing task for each active player.
     *
     * @param nextPlayers       The active players.
     * @param completionService The completion service.
     * @return The futures receiving the computation results.
     */
    private Map<Future<Optional<M>>, P> submitMoveComputingRequests(final Set<P> nextPlayers,
            final CompletionService<Optional<M>> completionService) {

        final Map<Future<Optional<M>>, P> futures = new LinkedHashMap<>(nextPlayers.size());
        for (final P nextPlayer : nextPlayers) {
            if (!this.strategies.containsKey(nextPlayer.getName())) {
                throw new IllegalStateException(String.format("State computed unknown next player %s.", nextPlayer));
            }

            final Strategy<P, S, M> strategy = this.strategies.get(nextPlayer.getName());
            final S stateCopy = this.state.deepCopy();
            futures.put(
                    completionService.submit(() -> strategy.computeNextMove(
                            this.id,
                            stateCopy.getPlayers().get(nextPlayer.getName()),
                            stateCopy,
                            this.maxComputationTimePerMove)),
                    nextPlayer);
        }
        return futures;
    }

    /**
     * Applies the next available move of the "fastest" player. If some move can be
     * successfully applied, {@link Observer#legalMoveApplied(Game, State, Player, Move)}
     * will be called, otherwise
     * {@link Observer#illegalMoveRejected(Game, State, Player, Optional, String)} will be
     * invoked for each illegal move returned until a legal move has been applied.
     * Pending moves or moves computed after a legal move of some player has been
     * applied are discarded without generating an event.
     *
     * @param completionService The completion service.
     * @param futures           The futures receiving the computation results.
     */
    private void applyNextPossibleMove(final CompletionService<Optional<M>> completionService,
            final Map<Future<Optional<M>>, P> futures) throws InterruptedException {

        Optional<P> playerDoingTheMove = Optional.empty();

        try {
            while (!futures.isEmpty()) {
                playerDoingTheMove = this.tryToApplyNextPossibleMove(completionService, futures);
                if (playerDoingTheMove.isPresent()) {
                    return;
                }
            }
        } finally {
            if (!futures.isEmpty()) {
                assert playerDoingTheMove.isPresent();

                for (final Map.Entry<Future<Optional<M>>, P> entry : futures.entrySet()) {
                    final Future<Optional<M>> future = entry.getKey();
                    future.cancel(true);

                    final S stateCopy = this.state.deepCopy();
                    final P overtakingPlayer = stateCopy.getPlayers().get(playerDoingTheMove.orElseThrow().getName());
                    final P overtakenPlayer = stateCopy.getPlayers().get(entry.getValue().getName());
                    this.callObservers((final Observer o) -> o.playerOvertaken(this, stateCopy, overtakenPlayer,
                            overtakingPlayer));
                }
            }
        }
    }

    /**
     * Tries to apply the next available move of the "fastest" player. If some move
     * can be successfully applied,
     * {@link Observer#legalMoveApplied(Game, State, Player, Move)} will be called,
     * otherwise {@link Observer#illegalMoveRejected(Game, State, Player, Optional, String)}
     * will be invoked for such an illegal move.
     *
     * @param completionService The completion service.
     * @param futures           The futures receiving the computation results.
     * @return The player for whom a legal move has been applied successfully (if
     *         any). If no legal move could be applied, an empty Optional is
     *         returned.
     */
    private Optional<P> tryToApplyNextPossibleMove(final CompletionService<Optional<M>> completionService,
            final Map<Future<Optional<M>>, P> futures) throws InterruptedException {

        final Future<Optional<M>> future = completionService.poll(this.maxComputationTimePerMove, TimeUnit.SECONDS);
        if (future == null) {
            // no strategy succeeded in finding a legal move within the configured time
            // window; choosing random moves
            for (final Map.Entry<Future<Optional<M>>, P> entry : futures.entrySet()) {
                final S stateCopy = this.state.deepCopy();
                final Optional<M> chosenMove = this
                        .chooseRandomMove(stateCopy.getPlayers().get(entry.getValue().getName()), stateCopy);
                if (chosenMove.isEmpty()) {
                    // No move available; this can happen if a previously chosen random move has won
                    // the game. In this case, we let the following strategies unpunished and do nothing.
                    continue;
                }

                this.handleOverdueMove(stateCopy.getPlayers().get(entry.getValue().getName()), chosenMove);
                try {
                    this.applyMoveIfPossible(stateCopy.getPlayers().get(entry.getValue().getName()), chosenMove.get());
                } catch (final GameException e) {
                    // the game itself did not succeed in finding a legal random move?!?
                    this.handleIllegalMove(stateCopy.getPlayers().get(entry.getValue().getName()), chosenMove,
                            e.getMessage());
                }
            }

            futures.clear();
            return Optional.empty();
        }

        final P playerDoingTheMove = futures.remove(future);
        Optional<M> move = Optional.empty();
        try {
            move = this.determineNextAvailableMove(future);
            if (move == null) {
                // strategy returned null which is not allowed, but we are going to condone it
                // for now
                move = Optional.empty();
            }
            if (move.isPresent()) {
                // check if strategy attempts to cheat by returning an unsupported custom move
                this.checkMove(move.get());

                this.applyMoveIfPossible(playerDoingTheMove, move.get());
                return Optional.of(playerDoingTheMove); // some legal move has been found and applied
            } else {
                // the player resigned the game
                playerDoingTheMove.setState(PlayerState.RESIGNED);
                final S stateCopy = this.state.deepCopy();
                this.callObservers((final Observer o) -> o.playerResigned(this, stateCopy,
                        stateCopy.getPlayers().get(playerDoingTheMove.getName())));
            }
        } catch (final GameException e) {
            // the strategy did not succeed in finding a legal move (or tried to cheat)
            this.handleIllegalMove(playerDoingTheMove, move, e.getMessage());
        }

        return Optional.empty();
    }

    /**
     * Handles an illegal move.
     *
     * @param player The player.
     * @param move   The move if present.
     * @param reason The reason why the move is illegal.
     */
    private void handleIllegalMove(final P player, final Optional<M> move, final String reason)
            throws InterruptedException {
        player.setState(PlayerState.LOST);
        final Optional<Move<?, ?>> moveTried = Optional.ofNullable(move.orElse(null));
        final S stateCopy = this.state.deepCopy();
        this.callObservers(
                (final Observer o) -> o.illegalMoveRejected(this, stateCopy,
                        stateCopy.getPlayers().get(player.getName()), moveTried, reason));
    }

    /**
     * Handles an overdue move.
     *
     * @param player     The player.
     * @param chosenMove The move that has been chosen.
     */
    private void handleOverdueMove(final P player, final Optional<M> chosenMove) throws InterruptedException {
        final Optional<Move<?, ?>> moveChosen = Optional.ofNullable(chosenMove.orElse(null));
        final S stateCopy = this.state.deepCopy();
        this.callObservers(
                (final Observer o) -> o.overdueMoveRejected(this, stateCopy,
                        stateCopy.getPlayers().get(player.getName()), moveChosen));
    }

    /**
     * Checks if the move is supported.
     *
     * @param move The move to check.
     * @throws GameException if the move is not supported.
     */
    private void checkMove(final M move) throws GameException {
        if (!this.moveChecker.check(move)) {
            throw new GameException(String.format("Unsupported move: %s.", move));
        }
    }

    /**
     * Returns the next available move from a {@link Future}.
     *
     * @param future The future.
     * @return The next available move returned by the strategy.
     * @throws GameException if the strategy caused an exception to be thrown.
     */
    private Optional<M> determineNextAvailableMove(final Future<Optional<M>> future)
            throws GameException, InterruptedException {
        try {
            return future.get();
        } catch (final ExecutionException e) {
            final Throwable cause = e.getCause();
            throw new GameException("The strategy did not succeed in finding a valid move: " + cause.getMessage(), e);
        }
    }

    /**
     * Applies a move for a given player to the current game state if possible. If
     * the move can be successfully applied,
     * {@link Observer#legalMoveApplied(Game, State, Player, Move)} will be called,
     * otherwise {@link Observer#illegalMoveRejected(Game, State, Player, Optional, String)}
     * will be invoked for each illegal move returned.
     *
     * @param player The current player.
     * @param move   The move to apply.
     * @throws GameException if the move could not be applied to the current game
     *                       state for some reason.
     */
    private void applyMoveIfPossible(final P player, final M move) throws GameException, InterruptedException {
        final S newState = this.state.deepCopy();
        move.applyTo(newState, newState.getPlayers().get(player.getName()));
        this.state = newState;

        final S stateCopy = this.state.deepCopy();
        this.callObservers((final Observer o) -> o.legalMoveApplied(this, stateCopy,
                stateCopy.getPlayers().get(player.getName()), move));
    }

    /**
     * Checks and adjusts the states of the players if necessary.
     */
    private void checkAndAdjustPlayerStatesIfNecessary() throws InterruptedException {
        final Collection<P> players = this.getPlayers().values();
        final List<P> playersPlaying = players.stream()
                .filter((final P player) -> player.getState().equals(PlayerState.PLAYING)).collect(Collectors.toList());
        final List<P> playersWon = players.stream()
                .filter((final P player) -> player.getState().equals(PlayerState.WON)).collect(Collectors.toList());

        final boolean gameOver;
        if (playersPlaying.isEmpty()) {
            // all players have stopped playing
            gameOver = true;
        } else if (!playersWon.isEmpty()) {
            // at least one player has won the game -- no time for losers (Queen)
            playersPlaying.forEach((final P player) -> player.setState(PlayerState.LOST));
            gameOver = true;
        } else if (playersPlaying.size() == 1) {
            // one player remains -- the winner takes them all (ABBA)
            playersPlaying.get(0).setState(PlayerState.WON);
            gameOver = true;
        } else {
            // there are at least two players participating at the game
            gameOver = false;
        }

        if (gameOver) {
            this.callObservers((final Observer o) -> o.finished(this, this.state.deepCopy()));
        }
    }
}
