/*
 * Copyright © 2020-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of gaming-core.
 *
 * Gaming-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Gaming-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * gaming-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.core.domain;

import java.util.List;

/**
 * A builder which allows to create a game.
 */
public interface GameBuilder {

    /**
     * The default maximum computation time per move in seconds.
     */
    int DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE = 5;

    /**
     * Changes the maximum computation time per move in seconds.
     * <p>
     * If this operation is called multiple times, only the last value will be retained. If not called, the default size
     * of {@link #DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE} seconds is used.
     *
     * @param newMaxComputationTimePerMove The new maximum computation time per move in seconds.
     * @return {@code this}
     */
    GameBuilder changeMaximumComputationTimePerMove(int newMaxComputationTimePerMove);

    /**
     * Adds {@link Observer}s to be attached to the game.
     *
     * @param newObservers The list of {@link Observer}s to be attached to the game.
     * @return {@code this}
     */
    GameBuilder addObservers(List<Observer> newObservers);

    /**
     * Builds the game. Can be called as many times as desired, each time returning a new game.
     *
     * @param id The ID of the game.
     * @throws GameException        if creating the game is not allowed by the rules of the game.
     * @throws InterruptedException if creating the game has been interrupted.
     */
    Game<?, ?, ?, ?> build(int id) throws GameException, InterruptedException;
}
